#include "lib.h"

#include "gui.h"
#include <cstddef>
#include <string>
using namespace std;



bool libLoaded = false;

const int gameSpeed = 240;
void timerHandler ();

void initLib () {
	allegro_init ();
	libLoaded = true;

	install_timer ();

	LOCK_VARIABLE ( fpsTimer );
	LOCK_VARIABLE ( gameTimer );
	LOCK_FUNCTION ( timerHandler );

	install_int_ex ( timerHandler, BPS_TO_TIMER ( gameSpeed ) );

	keyboard = new Keyboard ();
	#ifndef NOMOUSE
		try {
			mouse = new Mouse ();
		} catch ( Mouse::TooFewBtns ) {
			fatErrMsg ( "Sorry, your mouse has too few buttons." );
		}
	#endif

	try {
		setScreenRes ( fullscreen, 640, 480, 24 );
	} catch ( BuffError ) {
		fatErrMsg ( "Could not load fullscreen buffers." );
	} catch ( ResError ) {
		try {
			setScreenRes ( windowed, 640, 480, desktop_color_depth () );
		} catch ( BadRes err ) {
			fatErrMsg ( "Could not set video mode.  Cause:  " + err . what () );
		} catch ( BuffError ) { 
			fatErrMsg ( "Could not load windowed buffers." );
		} catch ( ResError ) {
		}
	}

	text_mode ( -1 );
}



// Error

void errMsg ( string msg ) {
	MsgBox ( "Error:\n  " + msg );
}

void fatErrMsg ( string msg ) {
	set_gfx_mode ( GFX_TEXT, 0, 0, 0, 0 );
	msg = "Fatal Error:\n   " + msg;
	allegro_message ( msg . c_str () );
	exit ( EXIT_FAILURE );
}



// Timer

volatile int fpsTimer;
volatile int gameTimer;

void timerHandler () {
	fpsTimer++;
	gameTimer++;
}
END_OF_FUNCTION ( timerHandler );

bool logicDone () {
	if ( gameTimer == 0 )
		return true;

	gameTimer--;
	return false;
}

void displayFPS () {
	static int fpsCount;
	static int framesCount;

	framesCount++;

	if ( fpsTimer >= gameSpeed ) {
		fpsCount = (gameSpeed * framesCount) / fpsTimer;
		fpsTimer = 0;
		framesCount = 0;
	}

	drawText ( numToText ( fpsCount ), 50, 50 );
}



// Keyboard

bool Keyboard::loaded = false;

Keyboard * keyboard;

Keyboard::Keyboard () {
	if ( ! libLoaded )
		fatErrMsg ( "Attempt to install keyboard before library is loaded." );
	if ( loaded )
		fatErrMsg ( "Attempt to install two keyboards." );

	install_keyboard ();
	loaded = true;
}

bool Keyboard::tap ( Lib_Key keyId ) {
	if ( press ( keyId ) ) {
		while ( press ( keyId ) ) { ; }
		return true;
	}
	return false;
}

Dir Keyboard::dir () {
	const Lib_Key u = key [ KEY_UP ];
	const Lib_Key r = key [ KEY_RIGHT ];
	const Lib_Key d = key [ KEY_DOWN ];
	const Lib_Key l = key [ KEY_LEFT ];

	if ( u && !r && !d && !l )
		return n;
	else if ( u && r && !d && !l )
		return ne;
	else if ( r && !u && !d && !l )
		return e;
	else if ( d && r && !u && !l )
		return se;
	else if ( d && !u && !r && !l )
		return s;
	else if ( d && l && !u && !r )
		return sw;
	else if ( l && !u && !d && !r )
		return w;
	else if ( u && l && ! d && ! r )
		return nw;
	else
		return nil;
}

void Keyboard::waitForTap ( Lib_Key keyID ) {
	while ( press ( keyID ) );
	while ( ! press ( keyID ) );
	while ( press ( keyID ) );
}

string Keyboard::textIn ( int maxChars ) {
	return string ();
}

int Keyboard::numIn ( int maxDigits ) {
	return 0;
}



// Mouse

#ifndef NOMOUSE

	bool Mouse::loaded = false;

	Mouse * mouse;

	Mouse::Mouse () throw ( Mouse::TooFewBtns ) {
		if ( ! libLoaded )
			fatErrMsg ( "Attempt to install mouse before library has been initialized." );
		if ( loaded )
			fatErrMsg ( "Attempt to install two mice." );

		if ( install_mouse () < MINBTNS ) {
			throw TooFewBtns ();
		}

		needPoll = mouse_needs_poll ();
		loaded = true;

		show_mouse ( screen );
	}

	bool Mouse::click () {
		for ( int i = 1; i <= MINBTNS; i++ )
			if ( mouse_b & i )
				return true;
		return false;
	}

	bool Mouse::inField ( int x1, int y1, int x2, int y2 ) {
		poll ();
		if ( mouse_x >= x1 && mouse_x <= x2 && mouse_y >= y1 && mouse_y <= y2 )
			return true;
		return false;
	}

#endif



// Graphics

enum { dblBuff, pageFlip } drawMode;

Lib_Pal pal;
Lib_Clr maskClr;

Bmp * screenBmp, * page1, * page2 = NULL;

void drawScreen () {
	switch ( drawMode ) {
		case pageFlip:
			show_video_bitmap ( screenBmp -> bmp );
			if ( screenBmp == page1 )
				screenBmp = page2;
			else
				screenBmp = page1;
		break; case dblBuff:
			screenBmp -> drawToScreen ( 0, 0 );
	}
}

void setScreenRes ( GfxMode mode, int w, int h, int clrDepth ) throw ( ResError ) {
	set_color_depth ( clrDepth );

	int Lib_gfxMode;

	switch ( mode ) {
		case fullscreen:
			Lib_gfxMode = GFX_AUTODETECT_FULLSCREEN;
			break;
		case windowed:
			Lib_gfxMode = GFX_AUTODETECT_WINDOWED;
			break;
	}

	if ( set_gfx_mode ( Lib_gfxMode, w, h, w, h * 2 ) != 0 )
		if ( set_gfx_mode ( Lib_gfxMode, w, h, 0, 0 ) != 0 )
			throw BadRes ( allegro_error );

	switch ( mode ) {
		case fullscreen:
			delete screenBmp;
			delete page1;
			delete page2;

			try {
				drawMode = pageFlip;

				page1 = new Bmp ( w, h, true );
				page2 = new Bmp ( w, h, true );
				screenBmp = page1;
			} catch ( Bmp::LoadError ) {
				drawMode = dblBuff;

				delete page1;
				delete page2;

				page1 = NULL;
				page2 = NULL;
				screenBmp = NULL;

				try {
					screenBmp = new Bmp ( w, h, false );
				} catch ( Bmp::LoadError err ) {
					throw BuffError ( "Could not create memory bitmap for fullscreen double-buffering.\nBitmap error: " + err . what () );
				}
			}
		break; case windowed:
			drawMode = dblBuff;
			try {
				screenBmp = new Bmp ( w, h, false );
			} catch ( Bmp::LoadError err ) {
				throw BuffError ( "Could not create memory bitmap for fullscreen double-buffering.\nBitmap error: " + err . what () );
			}
	}

	if ( set_display_switch_mode ( SWITCH_AMNESIA ) < 0 )
		throw BadSwitchMode ( "Could not set proper switch mode." );
	void ( * drawScreenPtr ) () = drawScreen;
	set_display_switch_callback ( SWITCH_IN, drawScreenPtr );

	if ( clrDepth == 8 )
		maskClr = 0;
	else
		maskClr = createClr ( 255, 0, 255 );
}



// Bmp

// Initialization/Cleanup

Bmp::Bmp ( int w, int h, bool video ) throw ( Bmp::LoadError ) {
	if ( ! libLoaded )
		fatErrMsg ( "Attempt to load " + numToText ( w ) + "x" + numToText ( h ) + " bitmap before library has been installed." );

	if ( video ) {
		bmp = create_video_bitmap ( w, h );
		if ( bmp == NULL )
			throw LoadError ( "Could not create " + numToText ( w ) + "x" + numToText ( h ) + " video bitmap." );
	} else {
		bmp = create_bitmap ( w, h );
		if ( bmp == NULL )
			throw LoadError ( "Could not create " + numToText ( w ) + "x" + numToText ( h ) + " memory bitmap." );
	}

	clearToClr ( maskClr );
}

Bmp::Bmp ( string bmpFile, bool video ) throw ( Bmp::LoadError ) {
	if ( ! libLoaded )
		fatErrMsg ( "Attempt to load bitmap from " + bmpFile + " before library has been installed." );

	if ( video ) {
		throw LoadError ( "Attempt to create video bitmap from a bitmap file: that's very bad, and I shouldn't have done it." );
	} else {
		bmp = load_bmp ( bmpFile . c_str (), pal );
		if ( bmp == NULL )
			throw LoadError ( "Could not create memory bitmap from file " + bmpFile + "." );
	}
}

Bmp::Bmp ( Lib_Bmp srcBmp, bool video ) throw ( Bmp::LoadError ) {
	if ( ! libLoaded )
		fatErrMsg ( "Attempt to load bitmap from library pointer before library has been installed." );

	if ( video ) {
		throw LoadError ( "Crazy stuff: attempt to create a video bitmap from a memory bitmap.  I shouldn't have done that!" );
	} else {
		bmp = srcBmp;
		if ( bmp == NULL )
			throw LoadError ( "Could not create memory bitmap from library bitmap." );
	}
}

Bmp::~Bmp () {
	destroy_bitmap ( bmp );
}

// Drawing

void Bmp::tile ( int targX1, int targY1, int targX2, int targY2, Bmp & targBmp ) {
	const int TARGW = targX2 - targX1 + 1;
	const int TARGH = targY2 - targY1 + 1;
	const int NUMROWS = int ( TARGH / retH () );
	const int NUMCOLS = int ( TARGW / retW () );

	for ( int row = 0; row < NUMROWS; row++ ) {
		for ( int col = 0; col < NUMCOLS; col++ ) {
			const int TARGX = targX1 + ( col * retW () );
			const int TARGY = targY1 + ( row * retH () );
			draw ( TARGX, TARGY, targBmp );
		}
		const int TARGX = targX1 + ( NUMCOLS * retW () );
		const int TARGY = targY1 + ( row * retH () );
		const int SRCX1 = 0;
		const int SRCY1 = 0;
		const int SRCX2 = TARGW - ( NUMCOLS * retW () ) - 1;
		const int SRCY2 = retH () - 1;
		drawPart ( TARGX, TARGY, SRCX1, SRCY1, SRCX2, SRCY2, targBmp );
	}
	for ( int col = 0; col < NUMCOLS; col++ ) {
		const int TARGX = targX1 + ( col * retW () );
		const int TARGY = targY1 + ( NUMROWS * retH () );
		const int SRCX1 = 0;
		const int SRCY1 = 0;
		const int SRCX2 = retW () - 1;
		const int SRCY2 = TARGH - ( NUMROWS * retH () ) - 1;
		drawPart ( TARGX, TARGY, SRCX1, SRCY1, SRCX2, SRCY2, targBmp );
	}
	const int TARGX = targX1 + ( NUMCOLS * retW () );
	const int TARGY = targY1 + ( NUMROWS * retH () );
	const int SRCX1 = 0;
	const int SRCY1 = 0;
	const int SRCX2 = TARGW - ( NUMCOLS * retW () ) - 1;
	const int SRCY2 = TARGH - ( NUMROWS * retH () ) - 1;
	drawPart ( TARGX, TARGY, SRCX1, SRCY1, SRCX2, SRCY2, targBmp );
}

void Bmp::tileMask ( int targX1, int targY1, int targX2, int targY2, Bmp & targBmp ) {
	const int TARGW = targX2 - targX1 + 1;
	const int TARGH = targY2 - targY1 + 1;

	int horzCntr;
	for ( horzCntr = 0; ( ( horzCntr + 1 ) * retW () ) < TARGW; horzCntr++ ) {
		int vertCntr;
		for ( vertCntr = 0; ( ( vertCntr + 1 ) * retH () ) < TARGH; vertCntr++ ) {
			const int BMPX = targX1 + horzCntr * retW () ;
			const int BMPY = targY1 + vertCntr * retH () ;
			drawMask ( BMPX, BMPY, targBmp );
		}
		const int TARGX = targX1 + horzCntr * retW ();
		const int TARGY = targY1 + horzCntr * retH ();
		const int SRCX1 = 0;
		const int SRCY1 = 0;
		const int SRCX2 = SRCX1 + retW () - 1;
		const int SRCY2 = TARGH - ( vertCntr * retH () );
		drawMaskPart ( TARGX, TARGY, SRCX1, SRCY1, SRCX2, SRCY2, targBmp );
	}
	int vertCntr;
	for ( vertCntr = 0; ( ( vertCntr + 1 ) * retH () ) < TARGH; vertCntr++ ) {
		const int TARGX = targX1 + horzCntr * retW ();
		const int TARGY = targY1 + horzCntr * retH ();
		const int SRCX1 = 0;
		const int SRCY1 = 0;
		const int SRCX2 = TARGW - ( horzCntr * retW () );
		const int SRCY2 = SRCY1 + retH () - 1;
		drawMaskPart ( TARGX, TARGY, SRCX1, SRCY1, SRCX2, SRCY2, targBmp );
	}
	const int TARGX = targX1 + horzCntr * retW ();
	const int TARGY = targY1 + horzCntr * retH ();
	const int SRCX1 = 0;
	const int SRCY1 = 0;
	const int SRCX2 = TARGW - ( horzCntr * retW () );
	const int SRCY2 = TARGH - ( vertCntr * retH () );
	drawPart ( TARGX, TARGY, SRCX1, SRCY1, SRCX2, SRCY2, targBmp );
}

// Collision

bool Bmp::coll ( int selfX, int selfY, int targX, int targY, Bmp & targ ) {

	const int selfX1 = selfX;
	const int selfY1 = selfY;
	const int selfX2 = selfX + retW () - 1;
	const int selfY2 = selfY + retH () - 1;

	const int targX1 = targX;
	const int targY1 = targY;
	const int targX2 = targX + targ . retW () - 1;
	const int targY2 = targY + targ . retH () - 1;

	if ( (selfX1 > targX2) || (selfX2 < targX1) || (selfY1 > targY2) || (selfY2 < targY1) )
		return false;

	const int selfCollX1 = (selfX1 > targX1) ? 0 : targX1 - selfX1;
	const int selfCollY1 = (selfY1 > targY1) ? 0 : targY1 - selfY1;
	const int selfCollX2 = (selfX2 < targX2) ? retW () - 1 : retW () - 1 - (selfX2 - targX2);
	const int selfCollY2 = (selfY2 < targY2) ? retH () - 1 : retH () - 1 - (selfY2 - targY2);

	const int targCollX1 = (targX1 > selfX1) ? 0 : selfX1 - targX1;
	const int targCollY1 = (targY1 > selfY1) ? 0 : selfY1 - targY1;
	const int targCollX2 = (targX2 < selfX2) ? targ . retW () - 1 : targ . retW () - 1 - (targX2 - selfX2);
	const int targCollY2 = (targY2 < selfY2) ? targ . retH () - 1 : targ . retH () - 1 - (targY2 - selfY2);

	for ( int selfRow = selfCollY1, targRow = targCollY1; selfRow <= selfCollY2; selfRow++, targRow++ )
		for ( int selfCol = selfCollX1, targCol = targCollX1; selfCol <= selfCollX2; selfCol++, targCol++ )
			if ( getpixel ( bmp, selfCol, selfRow ) != maskClr )
				if ( getpixel ( targ . bmp, targCol, targRow ) != maskClr )
					return true;
	return false;
}

void drawText ( string text, int x, int y ) {
	textout ( screenBmp -> bmp, font, text . c_str (), x, y, createClr ( 170, 170, 170 ) );
}
